import contextlib

from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django_audit_fields.admin import audit_fieldset_tuple
from edc_model_admin.history import SimpleHistoryAdmin
from rangefilter.filters import DateRangeFilterBuilder

from ...admin_site import edc_pharmacy_admin
from ...auth_objects import PHARMACIST_ROLE, PHARMACY_SUPER_ROLE
from ...exceptions import AllocationError, AssignmentError
from ...forms import StockForm
from ...models import RepackRequest, Stock, StockTransferItem
from ...utils import format_qty, get_related_or_none
from ..actions import (
    confirm_stock_from_queryset,
    go_to_add_repack_request_action,
    print_labels,
    print_stock_report_action,
)
from ..list_filters import (
    AllocationListFilter,
    ConfirmedListFilter,
    DecantedListFilter,
    HasOrderNumFilter,
    HasReceiveNumFilter,
    HasRepackNumFilter,
    ProductAssignmentListFilter,
    TransferredFilter,
)
from ..model_admin_mixin import ModelAdminMixin
from ..remove_fields_for_blinded_users import remove_fields_for_blinded_users


@admin.register(Stock, site=edc_pharmacy_admin)
class StockAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Stock"
    change_form_title = "Pharmacy: Stock"
    history_list_display = ()
    show_object_tools = False
    show_cancel = True
    list_per_page = 20

    show_form_tools = True
    show_history_label = True
    autocomplete_fields = ("container",)

    change_list_note = (
        "C=Confirmed, A=Allocated, T=Transferred to location, "
        "CL=Confirmed at location, D=Dispensed"
    )

    actions = (
        print_labels,
        confirm_stock_from_queryset,
        go_to_add_repack_request_action,
        print_stock_report_action,
    )

    form = StockForm

    ordering = ("code",)

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "stock_identifier",
                    "code",
                    "location",
                )
            },
        ),
        (
            "Product",
            {"fields": ("product", "container")},
        ),
        (
            "Batch",
            {"fields": ("lot",)},
        ),
        (
            "Quantity",
            {"fields": ("qty_in", "qty_out", "unit_qty_in", "unit_qty_out")},
        ),
        (
            "status",
            {
                "fields": (
                    "confirmed",
                    "allocation",
                    "in_transit",
                    "confirmed_at_location",
                    "stored_at_location",
                    "dispensed",
                    "destroyed",
                )
            },
        ),
        (
            "Receive / Repack",
            {
                "fields": (
                    "receive_item",
                    "repack_request",
                    "from_stock",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "formatted_code",
        "from_stock_changelist",
        "formulation",
        "formatted_confirmation",
        "formatted_allocation",
        "formatted_transferred",
        "formatted_confirmed_at_location",
        "formatted_dispensed",
        "location",
        "formatted_stored_at_location",
        "verified_assignment",
        "qty",
        "container_str",
        "unit_qty_in_out",
        "order_changelist",
        "receive_item_changelist",
        "repack_request_changelist",
        "stock_request_changelist",
        "allocation_changelist",
        "stock_transfer_item_changelist",
        "dispense_changelist",
        "created",
        "modified",
    )
    list_filter = (
        "location",
        "container__container_type",
        "container",
        ConfirmedListFilter,
        AllocationListFilter,
        TransferredFilter,
        "confirmed_at_location",
        "stored_at_location",
        "dispensed",
        ProductAssignmentListFilter,
        "product",
        "confirmation__confirmed_by",
        ("confirmation__confirmed_datetime", DateRangeFilterBuilder()),
        ("dispenseitem__dispense_item_datetime", DateRangeFilterBuilder()),
        HasOrderNumFilter,
        HasReceiveNumFilter,
        HasRepackNumFilter,
        DecantedListFilter,
        "created",
        "modified",
    )
    search_fields = (
        "stock_identifier",
        "from_stock__stock_identifier",
        "code",
        "lot__lot_no",
        "from_stock__code",
        "receive_item__id",
        "receive_item__receive__id",
        "receive_item__order_item__order__id",
        "repack_request__id",
        "allocation__registered_subject__subject_identifier",
        "allocation__stock_request_item__id",
        "allocation__stock_request_item__stock_request__id",
        "allocation__id",
        "stocktransferitem__stock_transfer__id",
        "dispenseitem__dispense__id",
    )
    readonly_fields = (
        "allocation",
        "code",
        "confirmation",
        "confirmed",
        "confirmed_at_location",
        "container",
        "destroyed",
        "dispensed",
        "from_stock",
        "in_transit",
        "location",
        "lot",
        "product",
        "qty_in",
        "qty_out",
        "receive_item",
        "repack_request",
        "stock_identifier",
        "stored_at_location",
        "unit_qty_in",
        "unit_qty_out",
    )

    def get_list_display(self, request):
        fields = super().get_list_display(request)
        return remove_fields_for_blinded_users(request, fields)

    def get_list_display_links(self, request, list_display):
        display_links = super().get_list_display_links(request, list_display)
        if not request.user.userprofile.roles.filter(
            name__in=[PHARMACIST_ROLE, PHARMACY_SUPER_ROLE]
        ).exists():
            with contextlib.suppress(ValueError):
                display_links.remove("formatted_code")
        return display_links

    def get_list_filter(self, request):
        fields = super().get_list_filter(request)
        return remove_fields_for_blinded_users(request, fields)

    def get_search_fields(self, request):
        fields = super().get_search_fields(request)
        return remove_fields_for_blinded_users(request, fields)

    def get_fieldsets(self, request, obj=None):
        fieldsets = super().get_fieldsets(request, obj)
        if not request.user.has_perm("edc_pharmacy.view_lot"):
            return [
                fieldset
                for fieldset in fieldsets
                if fieldset[0] and fieldset[0].lower() not in ["lot", "batch"]
            ]
        return fieldsets

    @admin.display(description="Assignment", ordering="lot__assignment__name")
    def verified_assignment(self, obj):
        try:
            obj.verify_assignment_or_raise()
        except AssignmentError:
            return format_html(
                "{}",
                mark_safe('<div style="color:red;">ERROR!</div>'),  # nosec B703, B308
            )
        except AllocationError:
            return mark_safe('<div style="color:red;">Allocation<BR>ERROR!</div>')
        return obj.lot.assignment

    @admin.display(description="Lot #", ordering="lot__lot_no")
    def formatted_lot(self, obj):
        return obj.lot.lot_no

    @admin.display(description="Stock #", ordering="-stock_identifier")
    def formatted_code(self, obj):
        return obj.code

    @admin.display(description="QTY", ordering="qty")
    def qty(self, obj):
        return format_qty(obj.qty_in - obj.qty_out, obj.container)

    @admin.display(description="Units", ordering="unit_qty_out")
    def unit_qty_in_out(self, obj):
        return format_qty(obj.unit_qty_in - obj.unit_qty_out, obj.container)

    @admin.display(description="STOCK #", ordering="-stock_identifier")
    def identifier(self, obj):
        return obj.stock_identifier.split("-")[0]

    @admin.display(description="C", boolean=True)
    def formatted_confirmation(self, obj):
        if obj:
            return bool(get_related_or_none(obj, "confirmation"))
        return None

    @admin.display(description="A", boolean=True)
    def formatted_allocation(self, obj):
        # if (
        #     obj
        #     and get_related_or_none(obj, "confirmation")
        #     and get_related_or_none(obj, "from_stock")
        # ):
        return bool(get_related_or_none(obj, "allocation"))
        # return None

    @admin.display(description="T", boolean=True)
    def formatted_transferred(self, obj):
        return obj.in_transit
        # if StockTransferItem.objects.filter(code=obj.code).exists():
        #     return True
        # return None

    @admin.display(description="CL", boolean=True)
    def formatted_confirmed_at_location(self, obj):
        return obj.confirmed_at_location
        # if (
        #     obj
        #     and get_related_or_none(obj, "confirmation")
        #     and get_related_or_none(obj, "from_stock")
        # ):
        #     confirmed = False
        #     if (
        #         stock_transfer_item := StockTransferItem.objects.filter(stock=obj)
        #         .order_by("transfer_item_datetime")
        #         .last()
        #     ):
        #         with contextlib.suppress(ObjectDoesNotExist):
        #             confirmed = bool(stock_transfer_item.confirmationatlocationitem)
        #     return confirmed
        # return None

    @admin.display(description="Bin")
    def formatted_stored_at_location(self, obj):
        if obj and obj.stored_at_location:
            label = obj.storagebinitem.storage_bin.bin_identifier
            url = reverse("edc_pharmacy_admin:edc_pharmacy_storagebin_changelist")
            url = f"{url}?q={label}"
            context = dict(
                url=url,
                label=label,
                title="Go to storage bin",
            )
            return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)
        return None

    @admin.display(description="D", boolean=True)
    def formatted_dispensed(self, obj):
        return obj.dispensed
        # if obj.location.name == CENTRAL_LOCATION:
        #     return None
        # return is_dispensed(obj)

    @admin.display(description="Container", ordering="container__name")
    def container_str(self, obj):
        return mark_safe("<BR>".join(str(obj.container).split(" ")))  # noqa: S308

    @admin.display(description="formulation", ordering="product__formulation__name")
    def formulation(self, obj):
        return obj.product.formulation

    @admin.display(description="assignment", ordering="product__assignment__name")
    def assignment(self, obj):
        if obj.product.assignment:
            return obj.product.assignment
        return None

    @admin.display(description="From stock #", ordering="from_stock__code")
    def from_stock_changelist(self, obj):
        if obj and get_related_or_none(obj, "from_stock"):
            url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
            url = f"{url}?q={obj.from_stock.code}"
            context = dict(
                url=url,
                label=obj.from_stock.code,
                title="Go to stock",
            )
            return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)
        return None

    @admin.display(description="Order #", ordering="-order__order_datetime")
    def order_changelist(self, obj):
        if obj.receive_item and obj.receive_item.order_item.order:
            url = reverse("edc_pharmacy_admin:edc_pharmacy_order_changelist")
            url = f"{url}?q={obj.receive_item.order_item.order.order_identifier}"
            context = dict(
                url=url,
                label=obj.receive_item.order_item.order.order_identifier,
                title="Go to order",
            )
            return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)
        return None

    @admin.display(
        description="Receive #", ordering="-receive_item__receive__receive_datetime"
    )
    def receive_item_changelist(self, obj):
        if obj and get_related_or_none(obj, "receive_item"):
            url = reverse("edc_pharmacy_admin:edc_pharmacy_receive_changelist")
            url = f"{url}?q={obj.receive_item.receive.receive_identifier}"
            context = dict(
                url=url,
                label=obj.receive_item.receive.receive_identifier,
                title="Go to receiving",
            )
            return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)
        return None

    @admin.display(description="Repack #", ordering="-repack_request__repack_datetime")
    def repack_request_changelist(self, obj):
        context = {}
        url = reverse("edc_pharmacy_admin:edc_pharmacy_repackrequest_changelist")
        if obj.repack_request:
            url = f"{url}?q={obj.repack_request.id}"
            context = dict(
                url=url,
                label=obj.repack_request.repack_identifier,
                title="Go to repack request",
            )
        elif RepackRequest.objects.filter(from_stock=obj).exists():
            url = f"{url}?q={obj.code}"
            context = dict(
                url=url,
                label="Repacks",
                title="Go to repack requests for this stock item",
            )
        if context:
            return render_to_string(
                "edc_pharmacy/stock/items_as_link.html",
                context=context,
            )
        return None

    @admin.display(
        description="Allocation #",
        ordering="-allocation__registered_subject__subject_identifier",
    )
    def allocation_changelist(self, obj):
        if obj and get_related_or_none(obj, "allocation"):
            url = reverse("edc_pharmacy_admin:edc_pharmacy_allocation_changelist")
            url = f"{url}?q={obj.code}"
            context = dict(
                url=url,
                label=obj.allocation.registered_subject.subject_identifier,
                title="Go to allocation",
            )
            return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)
        return None

    @admin.display(description="Request #")
    def stock_request_changelist(self, obj):
        if obj and get_related_or_none(obj, "allocation"):
            url = reverse("edc_pharmacy_admin:edc_pharmacy_stockrequest_changelist")
            url = f"{url}?q={obj.allocation.stock_request_item.stock_request.id}"
            context = dict(
                url=url,
                label=obj.allocation.stock_request_item.stock_request.request_identifier,
                title="Go to stock request",
            )
            return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)
        return None

    @admin.display(
        description="Transfer #",
        ordering="stock_transfer_item.to_location",
    )
    def stock_transfer_item_changelist(self, obj):
        if obj:
            try:
                stock_transfer_item_obj = obj.stocktransferitem_set.get(
                    stock_transfer__to_location=obj.location
                )
            except Stock.DoesNotExist:
                pass
            except StockTransferItem.DoesNotExist:
                pass
            else:
                url = reverse("edc_pharmacy_admin:edc_pharmacy_stocktransferitem_changelist")
                url = f"{url}?q={obj.code}"
                context = dict(
                    url=url,
                    label=stock_transfer_item_obj.stock_transfer.transfer_identifier,
                    title="Go to stock transfer item",
                )
                return render_to_string(
                    "edc_pharmacy/stock/items_as_link.html", context=context
                )
        return None

    @admin.display(description="DISPENSE #")
    def dispense_changelist(self, obj):
        if get_related_or_none(obj, "dispenseitem"):
            url = reverse("edc_pharmacy_admin:edc_pharmacy_dispense_changelist")
            url = f"{url}?q={obj.code}"
            context = dict(
                url=url,
                label=obj.dispenseitem.dispense.dispense_identifier,
                title="Go to dispense",
            )
            return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)
        return None



# Generated by Django 5.2.3 on 2025-06-27 13:02

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0072_stockout_bins"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="stockout",
            options={
                "default_permissions": ("view", "export", "viewallsites"),
                "verbose_name": "Stock availability",
                "verbose_name_plural": "Stock availability",
            },
        ),
    ]



# Generated by Django 4.2.11 on 2024-06-19 01:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_qareports", "0006_qareportlog"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="qareportlog",
            index=models.Index(fields=["accessed"], name="edc_qarepor_accesse_738ffe_idx"),
        ),
    ]



from django.db import models
from django.db.models import UniqueConstraint

from edc_model.models import BaseUuidModel, HistoricalRecords


class Manager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, name):
        return self.get(name)


class Medication(BaseUuidModel):
    name = models.CharField(max_length=35, unique=True)

    display_name = models.CharField(max_length=50, unique=True)

    notes = models.TextField(max_length=250, default="", blank=True)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return self.display_name

    def save(self, *args, **kwargs):
        self.name = self.name.strip().lower().replace(" ", "_")
        if not self.display_name:
            self.display_name = self.name.capitalize()
        super().save(*args, **kwargs)

    def natural_key(self):
        return self.name

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Medication"
        verbose_name_plural = "Medications"
        constraints = (
            UniqueConstraint(
                fields=["name", "display_name"],
                name="%(app_label)s_%(class)s_name_uniq",
            ),
        )



# Generated by Django 5.2.8 on 2025-11-20 23:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_metadata", "0034_alter_crfmetadata_options"),
        ("sites", "0002_alter_domain_unique"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="crfmetadata",
            index=models.Index(
                fields=["due_datetime"], name="edc_metadata_crfmetadata_a4idx"
            ),
        ),
        migrations.AddIndex(
            model_name="crfmetadata",
            index=models.Index(
                fields=["fill_datetime"], name="edc_metadata_crfmetadata_a5idx"
            ),
        ),
    ]



from .administration_view import AdministrationView
from .dashboard_view import DashboardView
from .error_handlers import edc_handler403, edc_handler404, edc_handler500
from .home_view import HomeView



from edc_auth.constants import (
    AUDITOR_ROLE,
    CLINICIAN_ROLE,
    CLINICIAN_SUPER_ROLE,
    NURSE_ROLE,
)
from edc_auth.site_auths import site_auths

from .auth_objects import OFFSTUDY, OFFSTUDY_SUPER, OFFSTUDY_VIEW, codenames


def update_site_auths():
    site_auths.add_group(*codenames, name=OFFSTUDY, no_delete=True)
    site_auths.add_group(*codenames, name=OFFSTUDY_SUPER)
    site_auths.add_group(*codenames, name=OFFSTUDY_VIEW, view_only=True)
    site_auths.update_role(OFFSTUDY, name=CLINICIAN_ROLE)
    site_auths.update_role(OFFSTUDY, name=NURSE_ROLE)
    site_auths.update_role(OFFSTUDY_SUPER, name=CLINICIAN_SUPER_ROLE)
    site_auths.update_role(OFFSTUDY_VIEW, name=AUDITOR_ROLE)


update_site_auths()



from __future__ import annotations

from typing import TYPE_CHECKING

from clinicedc_constants import PATIENT
from django.apps import apps as django_apps
from django.contrib import admin
from django.core.exceptions import ObjectDoesNotExist
from django_audit_fields import audit_fieldset_tuple

from edc_crf.admin import crf_status_fieldset_tuple
from edc_utils.date import to_local
from edc_visit_schedule.models import VisitSchedule
from edc_visit_tracking.utils import get_related_visit_model_cls

from ..choices import APPT_DATE_INFO_SOURCES

if TYPE_CHECKING:
    from edc_facility.models import HealthFacility

__all__ = ["NextAppointmentCrfModelAdminMixin"]


class NextAppointmentCrfModelAdminMixin(admin.ModelAdmin):
    fieldsets = (
        (None, {"fields": ("subject_visit", "report_datetime")}),
        (
            "Appointment",
            {
                "fields": (
                    "offschedule_today",
                    "appt_date",
                )
            },
        ),
        (
            "Appointment (Details)",
            {
                "fields": (
                    "visitschedule",
                    "health_facility",
                )
            },
        ),
        crf_status_fieldset_tuple,
        audit_fieldset_tuple,
    )

    radio_fields = {  # noqa: RUF012
        "offschedule_today": admin.VERTICAL,
        "crf_status": admin.VERTICAL,
        "info_source": admin.VERTICAL,
    }

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "visitschedule":
            visit_schedule_model_cls = django_apps.get_model(
                "edc_visit_schedule.visitschedule"
            )
            try:
                related_visit = self.related_visit(request)
            except ObjectDoesNotExist:
                kwargs["queryset"] = visit_schedule_model_cls.objects.none()
            else:
                kwargs["queryset"] = visit_schedule_model_cls.objects.filter(
                    visit_schedule_name=related_visit.visit_schedule_name,
                    schedule_name=related_visit.schedule_name,
                    active=True,
                )
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def formfield_for_choice_field(self, db_field, request, **kwargs):
        if db_field.name == "info_source":
            kwargs["choices"] = APPT_DATE_INFO_SOURCES
        return super().formfield_for_choice_field(db_field, request, **kwargs)

    def get_changeform_initial_data(self, request):
        initial = super().get_changeform_initial_data(request)
        # try to get next appointment
        try:
            related_visit = get_related_visit_model_cls().objects.get(
                id=initial.get("subject_visit")
            )
        except ObjectDoesNotExist:
            pass
        else:
            if next_appt := related_visit.appointment.next:
                initial.update(
                    info_source=self.get_default_info_source(request),
                    health_facility=self.get_default_health_facility(next_appt),
                    appt_date=to_local(next_appt.appt_datetime).date(),
                    appt_datetime=to_local(next_appt.appt_datetime),
                    visitschedule=VisitSchedule.objects.get(
                        visit_schedule_name=next_appt.visit_schedule_name,
                        schedule_name=next_appt.schedule_name,
                        visit_code=next_appt.visit_code,
                    ),
                )
        return initial

    def get_default_health_facility(self, next_appt) -> HealthFacility:
        return django_apps.get_model("edc_facility.healthfacility").objects.get(
            site=next_appt.site
        )

    def get_default_info_source(self, request):  # noqa: ARG002
        return django_apps.get_model("edc_appointment.infosources").objects.get(name=PATIENT)



from decimal import Decimal
from typing import Any


class DosageError(Exception):
    pass


class DosageCalculator:
    def __init__(
        self,
        dosage_guideline: Any | None = None,
        formulation: Any | None = None,
        weight_in_kgs: float | Decimal | None = None,
    ) -> None:
        self.dosage_guideline = dosage_guideline
        self.formulation = formulation
        if formulation.medication != dosage_guideline.medication:
            raise DosageError(
                "Medication mismatch. Guideline medication does not match formulation. "
                f"Got guideline.{dosage_guideline.medication} != formulation."
                f"{formulation.medication}."
            )
        if formulation.units.name != dosage_guideline.dose_units.name:
            raise DosageError(
                f"Invalid units. Guideline dose is in "
                f"'{dosage_guideline.dose_units}'. Got {formulation.units.name}."
            )
        if dosage_guideline.dose_per_kg and not weight_in_kgs:
            raise DosageError(
                "Expected weight_in_kgs for dosage_guideline dose in kgs. Got None"
            )
        self.weight_in_kgs = weight_in_kgs

    @property
    def dosage(self) -> float:
        """Returns the dosage as a float.

        The dosage is the dose in the units of dispensing, e.g 5, which
        might be expressed as 5 tabs 3 x per day

        The formulation_type and frequency of the dose depends on the formulation
        and dosage guideline.
        """
        return float(self.dose / float(self.strength))

    @property
    def strength(self):
        return self.formulation.strength

    @property
    def strength_units(self):
        return self.formulation.units.name

    @property
    def frequency(self):
        return self.dosage_guideline.frequency

    @property
    def dose_per_kg(self):
        return self.dosage_guideline.dose_per_kg

    @property
    def dose(self) -> float:
        """Returns the dose to be administered per day/hr or
        whatever the freq is.

        This is the guideline dose unless weight is factored in.

        """
        if self.weight_in_kgs and self.dose_per_kg:
            return float(self.dose_per_kg) * float(self.weight_in_kgs)
        return float(self.dosage_guideline.dose)



from django.contrib.auth.models import Group
from django.db import models
from django.utils.text import slugify

from edc_model.models import BaseUuidModel


class RoleModelManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)


class Role(BaseUuidModel):
    display_name = models.CharField(
        verbose_name="Display Name",
        max_length=250,
        unique=True,
        db_index=True,
        help_text="(suggest 40 characters max.)",
    )

    name = models.CharField(
        verbose_name="Name",
        max_length=250,
        unique=True,
        db_index=True,
        help_text="This is the stored value, required",
    )

    display_index = models.IntegerField(
        verbose_name="display index",
        default=0,
        db_index=True,
        help_text="Index to control display order if not alphabetical, not required",
    )

    objects = RoleModelManager()

    groups = models.ManyToManyField(Group)

    def __str__(self):
        return self.display_name

    def save(self, *args, **kwargs):
        if not self.name:
            self.name = slugify(self.display_name).lower()
        super().save(*args, **kwargs)

    def natural_key(self):
        return (self.name,)

    class Meta(BaseUuidModel.Meta):
        indexes = (
            *BaseUuidModel.Meta.indexes,
            models.Index(fields=["display_name", "display_index"]),
        )



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models


class HeartRateField(models.IntegerField):
    description = "Heart rate in BPM"

    def __init__(self, *args, **kwargs):
        if not kwargs.get("verbose_name"):
            kwargs["verbose_name"] = "Heart rate:"
        if not kwargs.get("validators"):
            kwargs["validators"] = [MinValueValidator(30), MaxValueValidator(200)]
        if not kwargs.get("help_text"):
            kwargs["help_text"] = "BPM"
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["verbose_name"]
        del kwargs["validators"]
        del kwargs["help_text"]
        return name, path, args, kwargs



# Generated by Django 4.0.4 on 2022-05-12 23:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_data_manager", "0021_auto_20210929_2343"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dataquery",
            name="recipients",
            field=models.ManyToManyField(
                blank=True,
                help_text="Select any additional recipients. Users in the `Site Data Manager` group are automatically included.",
                related_name="+",
                to="edc_data_manager.queryuser",
                verbose_name="Sent to",
            ),
        ),
        migrations.AlterField(
            model_name="queryrule",
            name="data_dictionaries",
            field=models.ManyToManyField(
                blank=True,
                help_text="select all that apply",
                related_name="+",
                to="edc_data_manager.crfdatadictionary",
                verbose_name="Question(s)",
            ),
        ),
        migrations.AlterField(
            model_name="queryrule",
            name="recipients",
            field=models.ManyToManyField(
                blank=True,
                help_text="Select any additional recipients. Users in the `Site Data Manager` group are automatically included.",
                related_name="+",
                to="edc_data_manager.queryuser",
                verbose_name="Send to",
            ),
        ),
        migrations.AlterField(
            model_name="queryrule",
            name="visit_schedule",
            field=models.ManyToManyField(
                blank=True,
                help_text="select all that apply",
                related_name="+",
                to="edc_data_manager.queryvisitschedule",
                verbose_name="Visit",
            ),
        ),
        migrations.AlterField(
            model_name="queryrule",
            name="visit_schedule_exclude",
            field=models.ManyToManyField(
                blank=True,
                help_text="select all that apply",
                related_name="+",
                to="edc_data_manager.queryvisitschedule",
                verbose_name="Visit (exclude)",
            ),
        ),
    ]



from edc_lab import ProcessingProfile, disposable, sputum, wb

blood_glucose_processing = ProcessingProfile(name="Blood Glucose", aliquot_type=wb)
chemistry_alt_processing = ProcessingProfile(name="Chem + ALT", aliquot_type=wb)
fbc_processing = ProcessingProfile(name="FBC", aliquot_type=wb)
hba1c_processing = ProcessingProfile(name="HbA1c", aliquot_type=wb)
insulin_processing = ProcessingProfile(name="Insulin", aliquot_type=wb)
lft_processing = ProcessingProfile(name="LFT", aliquot_type=wb)
lipids_processing = ProcessingProfile(name="Lipids", aliquot_type=wb)
poc_processing = ProcessingProfile(name="POC", aliquot_type=disposable)
rft_processing = ProcessingProfile(name="RFT", aliquot_type=wb)
sputum_processing = ProcessingProfile(name="Sputum", aliquot_type=sputum)
vl_processing = ProcessingProfile(name="Viral load", aliquot_type=wb)
cd4_processing = ProcessingProfile(name="CD4", aliquot_type=wb)



# Generated by Django 4.2.11 on 2024-08-15 13:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_qareports", "0014_alter_note_options_alter_note_status"),
    ]

    operations = [
        migrations.AlterField(
            model_name="note",
            name="status",
            field=models.CharField(
                choices=[
                    ("PENDING", "Data is pending"),
                    ("not_available", "Data not available"),
                ],
                max_length=25,
                null=True,
            ),
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.handlers.wsgi import WSGIRequest

if TYPE_CHECKING:
    from edc_crf.model_mixins import CrfModelMixin
    from edc_identifier.model_mixins import UniqueSubjectIdentifierModelMixin
    from edc_pdf_reports import CrfPdfReport

    class ModelMixin(CrfModelMixin, UniqueSubjectIdentifierModelMixin):
        pdf_report_cls = None


class PdfReportModelMixin:
    pdf_report_cls: type[CrfPdfReport]

    def get_pdf_report(self: ModelMixin, request: WSGIRequest):
        return self.pdf_report_cls(model_obj=self, request=request, user=request.user)

    class Meta:
        abstract = True



# Generated by Django 2.0 on 2018-01-03 11:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_locator", "0001_initial")]

    operations = [
        migrations.AlterModelOptions(
            name="subjectlocator", options={"verbose_name": "Subject Locator"}
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="may_visit_home",
            field=models.CharField(
                choices=[("Yes", "Yes"), ("No", "No")],
                max_length=25,
                verbose_name="Has the participant given permission for study staff <b>to make home visits</b> for follow-up purposes?",
            ),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="may_visit_home",
            field=models.CharField(
                choices=[("Yes", "Yes"), ("No", "No")],
                max_length=25,
                verbose_name="Has the participant given permission for study staff <b>to make home visits</b> for follow-up purposes?",
            ),
        ),
    ]



from django.http.response import HttpResponseRedirect
from django.urls.base import reverse

from ..requisition_verifier import RequisitionVerifier
from .base_requisition_view import BaseRequisitionView


class RequisitionVerifyActionsView(BaseRequisitionView):
    requisition_verifier_cls = RequisitionVerifier

    def post(self, request, *args, **kwargs):  # noqa: ARG002
        alert = 1
        error = 1
        subject_identifier = request.POST.get("subject_identifier")
        appointment_pk = request.POST.get("appointment")
        if request.POST.get("requisition_identifier"):
            requisition_identifier = (
                request.POST.get("requisition_identifier").replace("-", "").replace(" ", "")
            )
            requisition_verifier = self.requisition_verifier_cls(
                appointment_pk=appointment_pk,
                requisition_identifier=requisition_identifier,
            )
            if requisition_verifier.verified:
                alert = ""
                error = 0

        success_url = reverse(
            self.get_success_url(),
            kwargs=dict(
                subject_identifier=subject_identifier,
                appointment=appointment_pk,
                scanning=1,
                error=error,
            ),
        )
        return HttpResponseRedirect(
            redirect_to=f"{success_url}?alert={alert}#verify_requisition_identifier"
        )



import random
import re
from datetime import datetime
from zoneinfo import ZoneInfo

from django.conf import settings
from django.utils import timezone

safe_allowed_chars = "ABCDEFGHKMNPRTUVWXYZ2346789"


def get_safe_random_string(length=12, safe=None, allowed_chars=None):
    safe = True if safe is None else safe
    allowed_chars = allowed_chars or (
        "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRTUVWXYZ012346789!@#%^&*()?<>.,[]{}"
    )
    if safe:
        allowed_chars = "ABCDEFGHKMNPRTUVWXYZ2346789"
    return "".join([random.choice(allowed_chars) for _ in range(length)])  # nosec B311  # noqa: S311


def convert_php_dateformat(php_format_string):
    """Convert a date/datetime using a php format string
    as used by settings.SHORT_DATE_FORMAT.

    For example:
        obj.report_datetime.strftime(
            convert_php_dateformat(settings.SHORT_DATE_FORMAT)
        )
    """

    php_to_python = {
        "A": "%p",
        "D": "%a",
        "F": "%B",
        "H": "%H",
        "M": "%b",
        "N": "%b",
        "W": "%W",
        "Y": "%Y",
        "d": "%d",
        "e": "%Z",
        "h": "%I",
        "i": "%M",
        "l": "%A",
        "m": "%m",
        "s": "%S",
        "w": "%w",
        "y": "%y",
        "z": "%j",
        "j": "%d",
        "P": "%I:%M %p",
    }
    python_format_string = php_format_string
    for php, py in php_to_python.items():
        python_format_string = python_format_string.replace(php, py)
    return python_format_string


def convert_from_camel(name):
    """Converts from camel case to lowercase divided by underscores."""
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def formatted_datetime(
    dt: datetime | None,
    php_dateformat: str | None = None,
    tz: ZoneInfo | None = None,
    format_as_date: bool | None = None,
):
    """Returns a formatted datetime string, localized by default.

    format_as_date: does not affect the calculation, just the formatted output.
    """
    formatted = ""
    if dt:
        localized_dt = timezone.localtime(dt, timezone=tz)
        if format_as_date:
            php_dateformat = php_dateformat or settings.SHORT_DATE_FORMAT
            formatted = localized_dt.date().strftime(convert_php_dateformat(php_dateformat))
        else:
            php_dateformat = php_dateformat or settings.SHORT_DATETIME_FORMAT
            formatted = localized_dt.strftime(convert_php_dateformat(php_dateformat))
    return formatted


def formatted_date(dte, php_dateformat=None):
    """Returns a formatted datetime string."""
    if dte:
        php_dateformat = php_dateformat or settings.SHORT_DATE_FORMAT
        return dte.strftime(convert_php_dateformat(php_dateformat))
    return ""


def escape_braces(text: str) -> str:
    """Escapes text that may contain one or more braces
    (e.g., user supplied text) that is eventually passed to
    string.format() (where the inclusion of braces would and raise
    a ValueError)

    e.g.,
    `format_html(escape_braces("string with {braces} to escape"))`
    """
    return text.replace("{", "{{").replace("}", "}}")


def truncate_string(string: str, max_length: int) -> str:
    """Strips string of leading/trailing whitespace and truncates
    if > `max_length`.
    """
    if max_length < 1:
        raise ValueError("Max length must be >= 1")

    string = string.strip()
    if len(string) > max_length:
        return string[: max_length - 1].strip() + "…"
    return string



from .blood_results_fbg_form_validator_mixin import BloodResultsFbgFormValidatorMixin
from .blood_results_form_validator_mixin import BloodResultsFormValidatorMixin
from .blood_results_glu_form_validator_mixin import BloodResultsGluFormValidatorMixin



# Generated by Django 5.2.4 on 2025-09-10 03:24

import django_revision.revision_field
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_lab", "0034_alter_aliquot_site_alter_box_site_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="aliquot",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="box",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="boxitem",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="boxtype",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="consignee",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalaliquot",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalbox",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalboxitem",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalresult",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalresultitem",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalshipper",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="manifest",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="manifestitem",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="order",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="panel",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="result",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="resultitem",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="shipper",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
    ]



from django.db import models
from django.utils import timezone
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_sites.model_mixins import SiteModelMixin

from ..prescription import Rx
from .location import Location


class Manager(models.Manager):
    use_in_migrations = True


class Dispense(SiteModelMixin, BaseUuidModel):
    dispense_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    dispense_datetime = models.DateTimeField(default=timezone.now)

    dispensed_by = models.CharField(max_length=100, default="", blank=True)

    rx = models.ForeignKey(Rx, on_delete=models.PROTECT, null=True, blank=False)

    location = models.ForeignKey(Location, on_delete=models.PROTECT, null=True, blank=False)

    to_clinician = models.BooleanField(default=False)
    to_clinician_datetime = models.DateTimeField(null=True, blank=True)

    to_subject = models.BooleanField(default=False)
    to_subject_datetime = models.DateTimeField(null=True, blank=True)
    crf_label_lower = models.CharField(max_length=100, default="", blank=True)
    crf_field_name = models.CharField(max_length=100, default="", blank=True)
    crf_id = models.UUIDField(null=True, blank=True)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return self.dispense_identifier

    def save(self, *args, **kwargs):
        self.site = self.location.site
        if not self.dispense_identifier:
            self.dispense_identifier = f"{get_next_value(self._meta.label_lower):06d}"
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Dispense"
        verbose_name_plural = "Dispense"



from django.db import models

from ..utils import raise_if_offstudy


class OffstudyCrfModelMixin(models.Model):
    """Model mixin for CRF models.

    A mixin for CRF models to add the ability to determine
    if the subject is off study as of this CRFs report_datetime.

    CRFs by definition include CrfModelMixin in their declaration.
    See edc_visit_tracking.

    Also requires field "report_datetime"
    """

    def save(self, *args, **kwargs):
        self.raise_if_offstudy()
        super().save(*args, **kwargs)

    def raise_if_offstudy(self) -> None:
        if self.subject_identifier and self.report_datetime:
            raise_if_offstudy(
                source_obj=self,
                subject_identifier=self.subject_identifier,
                report_datetime=self.report_datetime,
            )

    class Meta:
        abstract = True



from clinicedc_constants import OTHER

LTFU_CHOICES = (
    ("unknown_address", "Changed to an unknown address"),
    ("never_returned", "Did not return despite reminders"),
    ("bad_contact_details", "Inaccurate contact details"),
    (OTHER, "Other, please specify ..."),
)



from django.db import models

from edc_constants.choices import YES_NO


class RequisitionVerifyModelMixin(models.Model):
    clinic_verified = models.CharField(max_length=15, choices=YES_NO, default="")

    clinic_verified_datetime = models.DateTimeField(null=True)

    class Meta:
        abstract = True



# Generated by Django 2.2.2 on 2019-07-25 16:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_data_manager", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="historicalqueryrule",
            name="rule_handler_name",
            field=models.CharField(
                choices=[
                    ("do_nothing", "Do Nothing"),
                    ("default", "Default"),
                    ("lumbar_puncture_q13", "Lumbar Puncture (Q13, 15, 21, 23, 24)"),
                ],
                default="default",
                max_length=150,
            ),
        ),
        migrations.AlterField(
            model_name="queryrule",
            name="rule_handler_name",
            field=models.CharField(
                choices=[
                    ("do_nothing", "Do Nothing"),
                    ("default", "Default"),
                    ("lumbar_puncture_q13", "Lumbar Puncture (Q13, 15, 21, 23, 24)"),
                ],
                default="default",
                max_length=150,
            ),
        ),
    ]



from clinicedc_constants import CANCELLED, CLOSED, OPEN
from django.db import models
from django.utils import timezone
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_registration.models import RegisteredSubject

from ...exceptions import InvalidContainer, StockRequestError
from ..medication import Formulation
from .container import Container
from .location import Location


class Manager(models.Manager):
    use_in_migrations = True


class StockRequest(BaseUuidModel):
    """A model to represent a stock request for subject stock.

    A request originates from, or is linked to, the research site
    using the location.
    """

    request_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    request_datetime = models.DateTimeField(default=timezone.now)

    start_datetime = models.DateTimeField(
        default=None,
        null=True,
        blank=True,
        help_text=(
            "If provided, will not include appointments before this date. May be left blank"
        ),
    )

    cutoff_datetime = models.DateTimeField(
        default=None,
        null=True,
        blank=True,
        help_text=(
            "If provided, will not include appointments after this date. May be left blank"
        ),
    )

    location = models.ForeignKey(
        Location,
        verbose_name="Requested from",
        on_delete=models.PROTECT,
        null=True,
        blank=False,
        limit_choices_to={"site_id__isnull": False},
    )

    formulation = models.ForeignKey(
        Formulation,
        on_delete=models.PROTECT,
        null=True,
        blank=False,
    )

    container = models.ForeignKey(
        Container,
        on_delete=models.PROTECT,
        null=True,
        blank=False,
        limit_choices_to={"may_dispense_as": True},
    )

    containers_per_subject = models.PositiveSmallIntegerField(
        verbose_name="Number of containers per subject", default=3
    )

    item_count = models.IntegerField(
        verbose_name="Item count",
        default=0,
        help_text="Matches the number of Request items.",
    )

    subject_identifiers = models.TextField(
        verbose_name="Include ONLY these subjects in this request. (Usually left blank)",
        default="",
        blank=True,
        help_text=(
            "By adding subject identifiers in this box, only these subjects "
            "will be included in the request. All others will be ignored."
        ),
    )

    excluded_subject_identifiers = models.TextField(
        verbose_name="Exclude these subjects from this request. (Usually left blank)",
        default="",
        blank=True,
    )

    labels = models.TextField(
        verbose_name="Labels",
        default="",
        blank=True,
        help_text=(
            "A cell to capture and confirm printed/scanned labels related to this "
            "Stock request. See StockRequestItem."
        ),
    )

    cancel = models.CharField(
        verbose_name="To cancel this request, type the word 'CANCEL' here and save the form:",
        max_length=6,
        default="",
        blank=True,
    )
    status = models.CharField(
        max_length=25,
        choices=(
            (OPEN, OPEN.title()),
            (CLOSED, CLOSED.title()),
            (CANCELLED, CANCELLED.title()),
        ),
        default=OPEN,
    )

    task_id = models.UUIDField(null=True)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return self.request_identifier

    def save(self, *args, **kwargs):
        if not self.request_identifier:
            next_id = get_next_value(self._meta.label_lower)
            self.request_identifier = f"{next_id:06d}"
        if not self.container.may_request_as:
            raise InvalidContainer(
                "Invalid stock.container. Must be a `subject-specific` container. "
                "Perhaps catch this in the form."
            )
        if not self.formulation:
            raise StockRequestError("Formulation may not be null")

        if self.subject_identifiers:
            subject_identifiers = self.subject_identifiers.split("\n")
            subject_identifiers = [s.strip() for s in subject_identifiers]
            self.subject_identifiers = "\n".join(subject_identifiers)
            if RegisteredSubject.objects.values("subject_identifier").filter(
                subject_identifier__in=subject_identifiers
            ).count() != len(subject_identifiers):
                raise StockRequestError(
                    "Invalid subject_identifier listed. Perhaps catch this in the form"
                )

        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Stock request"
        verbose_name_plural = "Stock requests"



# Generated by Django 5.2.7 on 2025-10-24 04:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_data_manager", "0042_alter_datadictionary_revision_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalqueryrule",
            name="comment",
            field=models.TextField(blank=True, default=""),
        ),
        migrations.AlterField(
            model_name="historicalqueryrule",
            name="query_text",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Generic query text for auto-generated queries.",
            ),
        ),
        migrations.AlterField(
            model_name="historicalqueryrule",
            name="reference_model",
            field=models.CharField(default="", editable=False, max_length=150),
        ),
        migrations.AlterField(
            model_name="queryrule",
            name="comment",
            field=models.TextField(blank=True, default=""),
        ),
        migrations.AlterField(
            model_name="queryrule",
            name="query_text",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Generic query text for auto-generated queries.",
            ),
        ),
        migrations.AlterField(
            model_name="queryrule",
            name="reference_model",
            field=models.CharField(default="", editable=False, max_length=150),
        ),
    ]



from django.contrib import messages
from django.utils.translation import gettext as _


def get_message_text(level: int) -> str:
    if level == messages.WARNING:
        return _(
            "You have permissions to view forms and data from sites other than the current. "
        )
    if level == messages.ERROR:
        return _(
            "Showing data from the current site only. Although you have permissions to view "
            "data from multiple sites you also have permissions to add, change or delete "
            "data. This is not permitted when viewing data from multiple sites. Contact your "
            "system administrator."
        )
    return ""



from .edc_permissions import EdcPermissions
from .medication import (
    Assignment,
    DosageGuideline,
    Formulation,
    FormulationType,
    FrequencyUnits,
    Medication,
    Route,
    Units,
)
from .prescription import Rx, RxRefill
from .proxy_models import SiteProxy, VisitSchedule
from .reports import StockAvailability
from .scan_duplicates import ScanDuplicates
from .signals import (
    create_or_update_refills_on_post_save,
    dispense_item_on_post_delete,
    dispense_item_on_post_save,
    order_item_on_post_save,
    receive_item_on_post_delete,
    receive_item_on_post_save,
    receive_on_post_save,
    repack_request_on_post_save,
    stock_adjustment_on_post_save,
    stock_on_post_delete,
    stock_on_post_save,
    stock_request_item_on_post_save,
)
from .stock import (
    Allocation,
    AllocationProxy,
    Confirmation,
    ConfirmationAtLocation,
    ConfirmationAtLocationItem,
    Container,
    ContainerType,
    ContainerUnits,
    Dispense,
    DispenseItem,
    Location,
    Lot,
    Order,
    OrderItem,
    Product,
    Receive,
    ReceiveItem,
    RepackRequest,
    Stock,
    StockAdjustment,
    StockProxy,
    StockRequest,
    StockRequestItem,
    StockTransfer,
    StockTransferItem,
    StorageBin,
    StorageBinItem,
    Supplier,
)
from .storage import (
    Box,
    Room,
    Shelf,
    UnitType,
    get_location,
    get_room,
    get_shelf,
    repackage,
    repackage_for_subject,
)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.core.handlers.wsgi import WSGIRequest
from django.db import transaction
from django.utils import timezone

from ..constants import CENTRAL_LOCATION
from ..exceptions import StockTransferError
from ..utils import is_dispensed

if TYPE_CHECKING:
    from ..models import (
        ConfirmationAtLocationItem,
        Stock,
        StockTransfer,
        StockTransferItem,
        StorageBinItem,
    )


def transfer_stock_to_location(
    stock_transfer: StockTransfer, stock_codes: list[str], request: WSGIRequest = None
) -> tuple[list[str], list[str], list[str], list[str]]:
    stock_model_cls: type[Stock] = django_apps.get_model("edc_pharmacy.stock")
    stock_transfer_item_model_cls: type[StockTransferItem] = django_apps.get_model(
        "edc_pharmacy.stocktransferitem"
    )
    storage_bin_item_model_cls: type[StorageBinItem] = django_apps.get_model(
        "edc_pharmacy.storagebinitem"
    )
    confirmation_at_location_item_model_cls: type[ConfirmationAtLocationItem] = (
        django_apps.get_model("edc_pharmacy.confirmationatlocationitem")
    )
    transferred, dispensed_codes, skipped_codes, invalid_codes = [], [], [], []
    for stock_code in stock_codes:
        if not stock_model_cls.objects.filter(code=stock_code).exists():
            invalid_codes.append(stock_code)
            continue
        # must be confirmed, allocated and at the "from" location
        opts = dict(
            code=stock_code,
            confirmation__isnull=False,
            allocation__registered_subject__isnull=False,
            location=stock_transfer.from_location,
        )
        try:
            stock_model_cls.objects.get(**opts)
        except ObjectDoesNotExist:
            skipped_codes.append(stock_code)
        else:
            if stock_transfer.to_location.name == CENTRAL_LOCATION:
                opts.update(
                    allocation__registered_subject__site=stock_transfer.from_location.site,
                )
            else:
                opts.update(
                    allocation__registered_subject__site=stock_transfer.to_location.site
                )
            try:
                stock_obj = stock_model_cls.objects.get(**opts)
            except ObjectDoesNotExist:
                skipped_codes.append(stock_code)
            else:
                if is_dispensed(stock_code):
                    dispensed_codes.append(stock_code)
                else:
                    with transaction.atomic():
                        # get or create stock_transfer_item and relate
                        # to this stock transfer
                        stock_transfer_item_model_cls.objects.create(
                            stock=stock_obj,
                            stock_transfer=stock_transfer,
                            user_created=request.user.username,
                            created=timezone.now(),
                        )

                        # remove stock from the storage bin, if stored at
                        # a site bin
                        storage_bin_item_model_cls.objects.filter(stock=stock_obj).delete()
                        stock_obj.stored_at_location = False

                        # delete confirmation (you can still see it in history)
                        confirmation_at_location_item_model_cls.objects.filter(
                            stock=stock_obj
                        ).delete()

                        # change location of stock
                        stock_obj.location = stock_transfer.to_location

                        # save
                        stock_obj.save()

                        transferred.append(stock_code)

                        if len(stock_codes) != (
                            len(transferred)
                            + len(dispensed_codes)
                            + len(skipped_codes)
                            + len(invalid_codes)
                        ):
                            raise StockTransferError(
                                "Some codes were not accounted for. Cancelling transfer"
                            )
    return transferred, dispensed_codes, skipped_codes, invalid_codes


__all__ = ["transfer_stock_to_location"]



from __future__ import annotations

from django.contrib import admin
from django_audit_fields.admin import ModelAdminAuditFieldsMixin


class TabularInlineMixin(ModelAdminAuditFieldsMixin, admin.TabularInline):
    insert_before_fieldset: str | None = "Audit"

    template = "edc_model_admin/admin/tabular.html"

    def get_formset(self, request, obj=None, **kwargs):
        formset = super().get_formset(request, obj=None, **kwargs)
        formset.insert_before_fieldset = self.insert_before_fieldset
        return formset



from clinicedc_constants import CLOSED, NEW, OPEN
from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone

from edc_notification.notification import Notification

NOTIFY_ON_NEW_AND_NO_REFERENCE_OBJ = "notify_on_new_and_no_reference_obj"
NOTIFY_ON_NEW = "notify_on_new"
NOTIFY_ON_OPEN = "notify_on_open"
NOTIFY_ON_CLOSE = "notify_on_closed"
NOTIFY_ON_CHANGED_REFERENCE_OBJ = "notify_on_changed_reference_obj"


class ActionItemNotification(Notification):
    """A Notification class for the action item model.

    This class is the default superclass for an Action's
    notification class.

    See `Action.notification_cls()`

    * Sends an initial notification when an action or action_item
      is created (but only if the action item is created before
      the reference object is created);

    * Send an update notification if the reference object is changed.
      The reference object is considered changed if one of the values
      for a field listed in `notification_fields` is changed.

    """

    # set by default
    notification_action_name: str = None
    # set by class user, list of field(s) from
    # the action_item.reference_obj
    notification_fields: list[str] = []

    # TODO: not sure what this does?
    notify_on_new: bool = False

    # action_item created before the reference obj
    notify_on_new_and_no_reference_obj: bool = True
    # reference obj creates action_item and status set to OPEN
    notify_on_open: bool = False
    # action_item status set to CLOSED
    notify_on_closed: bool = False
    # reference obj changed, action item status set to OPEN
    notify_on_changed_reference_obj: bool = True
    # set by default to the Action's reference_model
    model: str = None

    email_subject_template: str = (
        "{updated_subject_line}{test_subject_line}{protocol_name}: "
        "{display_name} "
        "for {subject_identifier}"
    )
    email_body_template_new: str = (
        "\n\nDo not reply to this email\n\n"
        "{test_body_line}"
        'A report titled "{parent_reference_verbose_name}" has been submitted '
        "for patient {subject_identifier} "
        "at the {site_name} site which requires additional information.\n\n"
        "Please complete the following report:\n\n"
        "Title: {display_name}.\n\n"
        "You received this message because you are subscribed to receive these "
        "notifications in your user profile.\n\n"
        "{test_body_line}"
        "Thanks."
    )
    email_body_template_update: str = (
        "\n\nDo not reply to this email\n\n"
        "{test_body_line}"
        "An updated report has been submitted.\n\n"
        'The information related to the report "{display_name}" for patient '
        "{subject_identifier} "
        "at the {site_name} site has changed. You may wish to review.\n\n"
        "You received this message because you are subscribed to receive these "
        "notifications in your user profile.\n\n"
        "{test_body_line}"
        "Thanks."
    )
    email_footer_template: str = (
        "\n\n-----------------\n"
        'To unsubscribe remove "{display_name}" from your chosen '
        "email notifications in your user profile.\n\n"
        "{name}\n"
        "{notification_reason}:{pk}\n"
        "{message_datetime} (UTC)"
    )

    def __init__(self):
        super().__init__()
        self.updated_subject_line = ""
        self.updated_body_line = ""

    def notify_on(self, action_item=None, reference_obj=None):
        updated = False
        if (
            self.notify_on_new_and_no_reference_obj
            and action_item.status == NEW
            and not reference_obj
        ):
            notify = NOTIFY_ON_NEW_AND_NO_REFERENCE_OBJ
        elif self.notify_on_new and action_item.status == NEW and reference_obj:
            notify = NOTIFY_ON_NEW
        elif self.notify_on_open and action_item.status == OPEN and reference_obj:
            notify = NOTIFY_ON_OPEN
        elif self.notify_on_closed and action_item.status == CLOSED and reference_obj:
            notify = NOTIFY_ON_CLOSE
        elif (
            self.notify_on_changed_reference_obj
            and action_item.status != NEW
            and reference_obj
        ):
            notify = NOTIFY_ON_CHANGED_REFERENCE_OBJ
            updated = True
        else:
            notify = False
        return notify, updated

    def notify(self, force_notify=None, use_email=None, use_sms=None, **kwargs):
        """Overridden to only call `notify` if action name matches."""
        notified = False
        instance = kwargs.get("instance")
        if instance._meta.label_lower == "edc_action_item.actionitem":
            try:
                action_name = instance.action_type.name
            except AttributeError:
                pass
            else:
                if action_name == self.notification_action_name:
                    action_item = kwargs.get("instance")
                    reference_obj = self.get_reference_obj(action_item)

                    notify, updated = self.notify_on(
                        action_item=action_item, reference_obj=reference_obj
                    )

                    kwargs.update(updated=updated)
                    kwargs.update(instance=reference_obj)
                    kwargs.update(action_item=action_item)

                    if notify:
                        email_body_template = (
                            self.email_body_template_update
                            if updated
                            else self.email_body_template_new
                        )
                        kwargs.update(
                            notification_reason=notify,
                            updated=updated,
                            parent_reference_verbose_name=(
                                action_item.reference_model_cls()._meta.verbose_name
                            ),
                            subject_identifier=action_item.subject_identifier,
                            site_name=action_item.site.name,
                            pk=action_item.pk,
                        )
                        notified = super().notify(
                            force_notify=force_notify,
                            use_email=use_email,
                            use_sms=use_sms,
                            email_body_template=email_body_template,
                            **kwargs,
                        )
        return notified

    def notify_on_condition(self, **kwargs):
        notify_on_condition = False
        updated = kwargs.get("updated")
        reference_obj = kwargs.get("instance")
        action_item = kwargs.get("action_item")
        if not reference_obj:
            emailed = action_item.emailed
        else:
            emailed = reference_obj.action_item.emailed
        if not emailed and not updated:
            self.updated_subject_line = ""
            self.updated_body_line = "A report"
            notify_on_condition = True
        elif updated:
            self.updated_subject_line = "*UPDATE* "
            self.updated_body_line = "An updated report"
            notify_on_condition = True
        return notify_on_condition

    @staticmethod
    def get_reference_obj(action_item):
        try:
            reference_obj = action_item.reference_obj
        except ObjectDoesNotExist:
            reference_obj = None
        return reference_obj

    @staticmethod
    def get_action_name(instance):
        """Returns that action_name."""
        try:
            action_name = instance.action_item.action_cls.name
        except AttributeError:
            try:
                action_name = instance.action_type.name
            except AttributeError:
                action_name = None
        return action_name

    def get_template_options(self, **kwargs):
        template_options = super().get_template_options(**kwargs)
        template_options.update(
            updated_subject_line=self.updated_subject_line,
            updated_body_line=self.updated_body_line,
        )
        return template_options

    def post_notification_actions(self, email_sent=None, **kwargs):
        """Record the datetime of first email sent."""
        action_item = kwargs.get("action_item")
        if email_sent and not action_item.emailed:
            action_item.emailed = True
            action_item.emailed_datetime = timezone.now()
            action_item.save_without_historical_record()



# Generated by Django 2.1.7 on 2019-03-04 23:23

import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [("edc_lab", "0018_auto_20190201_0446")]

    operations = [
        migrations.AlterField(
            model_name="aliquot",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="aliquot",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="aliquot",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="aliquot",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="aliquot",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="box",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="box",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="box",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="box",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="box",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="boxitem",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="boxitem",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="boxitem",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="boxitem",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="boxitem",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="boxtype",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="boxtype",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="boxtype",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="boxtype",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="boxtype",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="consignee",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="consignee",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="consignee",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="consignee",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="consignee",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalaliquot",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalaliquot",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalaliquot",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalaliquot",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalaliquot",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalbox",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalbox",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalbox",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalbox",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalbox",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalboxitem",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalboxitem",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalboxitem",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalboxitem",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalboxitem",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalresult",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalresult",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalresult",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalresult",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalresult",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalresultitem",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalresultitem",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalresultitem",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalresultitem",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalresultitem",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalshipper",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalshipper",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalshipper",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalshipper",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalshipper",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="manifest",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="manifest",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="manifest",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="manifest",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="manifest",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="manifestitem",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="manifestitem",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="manifestitem",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="manifestitem",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="manifestitem",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="order",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="order",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="order",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="order",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="order",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="panel",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="panel",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="panel",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="panel",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="panel",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="result",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="result",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="result",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="result",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="result",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="resultitem",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="resultitem",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="resultitem",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="resultitem",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="resultitem",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="shipper",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="shipper",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="shipper",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="shipper",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="shipper",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
    ]



from .chol_initial_review_model_mixin import CholInitialReviewModelMixin
from .hiv_initial_model_mixins import (
    HivArvInitiationModelMixin,
    HivArvMonitoringModelMixin,
)
from .ncd_initial_review_model_mixin import NcdInitialReviewModelMixin



from django.conf import settings

from edc_dashboard.middleware_mixins import EdcTemplateMiddlewareMixin
from edc_subject_dashboard.dashboard_urls import dashboard_urls

from .dashboard_templates import dashboard_templates


class DashboardMiddleware(EdcTemplateMiddlewareMixin):
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        self.check_for_required_request_attrs(request)
        return self.get_response(request)

    def process_view(self, request, *args):
        """Adds/Updates references to templates."""
        request.url_name_data.update(**dashboard_urls)
        template_data = getattr(settings, "SUBJECT_DASHBOARD_BASE_TEMPLATES", {})
        template_data.update(**dashboard_templates)
        request.template_data.update(**template_data)

    def process_template_response(self, request, response):
        if getattr(response, "context_data", None):
            response.context_data.update(**request.url_name_data)
            response.context_data.update(**request.template_data)
        return response



from django.conf import settings
from reportlab.graphics.shapes import Group, String
from reportlab.lib import colors


def draw_label_watermark(label, width, height, **string_options):
    if word := getattr(settings, "EDC_PHARMACY_LABEL_WATERMARK_WORD", None):
        string_opts = dict(
            fontName="Helvetica",
            fontSize=28,
            textAnchor="middle",
            fillColor=colors.Color(0.5, 0.5, 0.5, alpha=0.7),
        )
        string_opts.update(string_options)
        text_group = Group()
        watermark = String(height / 2, 10, word, **string_opts)
        text_group.add(watermark)
        text_group.translate(width / 3, height - height * 0.95)
        text_group.rotate(45)
        label.add(text_group)



# Generated by Django 2.1 on 2018-08-09 03:03

import _socket
import django.db.models.deletion
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
from django.db import migrations, models

import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [("edc_action_item", "0007_auto_20180707_1715")]

    operations = [
        migrations.CreateModel(
            name="Reference",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("subject_identifier", models.CharField(max_length=50)),
                (
                    "parent_reference_identifier",
                    models.CharField(max_length=30, null=True),
                ),
                (
                    "related_reference_identifier",
                    models.CharField(max_length=30, null=True),
                ),
                ("action_identifier", models.CharField(max_length=25, unique=True)),
                (
                    "report_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                (
                    "action_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="action",
                        to="edc_action_item.ActionType",
                        verbose_name="Action",
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.RemoveField(model_name="action", name="action_type"),
        migrations.AlterField(
            model_name="actionitem",
            name="linked_to_reference",
            field=models.BooleanField(
                default=False,
                editable=False,
                help_text='True if this action is linked to it\'s reference_model.Initially False if this action is created before reference_model.Always True when reference_model creates the action.Set to True when reference_model is created and "links" to this action.(Note: reference_model looks for actions where linked_to_reference is False before attempting to create a new ActionItem).',
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="linked_to_reference",
            field=models.BooleanField(
                default=False,
                editable=False,
                help_text='True if this action is linked to it\'s reference_model.Initially False if this action is created before reference_model.Always True when reference_model creates the action.Set to True when reference_model is created and "links" to this action.(Note: reference_model looks for actions where linked_to_reference is False before attempting to create a new ActionItem).',
            ),
        ),
        migrations.AlterUniqueTogether(name="actionitem", unique_together=set()),
        migrations.DeleteModel(name="Action"),
    ]



from typing import Any

from clinicedc_constants import COMPLETE, INCOMPLETE
from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist


def update_crf_status_for_instance(instance: Any) -> None:
    """Only works for CRFs, e.g. have related_visit."""
    if hasattr(instance, "related_visit"):
        crf_status_model_cls = django_apps.get_model("edc_crf.crfstatus")
        opts = dict(
            subject_identifier=instance.related_visit.subject_identifier,
            visit_schedule_name=instance.related_visit.visit_schedule_name,
            schedule_name=instance.related_visit.schedule_name,
            visit_code=instance.related_visit.visit_code,
            visit_code_sequence=instance.related_visit.visit_code_sequence,
            label_lower=instance._meta.label_lower,
        )
        if instance.crf_status == COMPLETE:
            crf_status_model_cls.objects.filter(**opts).delete()
        elif instance.crf_status == INCOMPLETE:
            try:
                crf_status_model_cls.objects.get(**opts)
            except ObjectDoesNotExist:
                crf_status_model_cls.objects.create(**opts)



from edc_dashboard.view_mixins import EdcViewMixin
from edc_listboard.view_mixins import ListboardFilterViewMixin, SearchFormViewMixin
from edc_listboard.views import ListboardView
from edc_navbar import NavbarViewMixin

from ...view_mixins import FormActionViewMixin


class BaseListboardView(
    EdcViewMixin,
    FormActionViewMixin,
    SearchFormViewMixin,
    NavbarViewMixin,
    ListboardFilterViewMixin,
    ListboardView,
):
    navbar_name = "specimens"



from django.contrib.admin.decorators import register
from django_audit_fields.admin import audit_fieldset_tuple

from edc_model_admin.history import SimpleHistoryAdmin

from ..admin_site import edc_visit_schedule_admin
from ..models import VisitSchedule
from ..site_visit_schedules import site_visit_schedules


@register(VisitSchedule, site=edc_visit_schedule_admin)
class VisitScheduleAdmin(SimpleHistoryAdmin):
    actions = ("populate_visit_schedule",)

    fieldsets = (
        [
            None,
            {
                "fields": (
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_name",
                    "timepoint",
                    "active",
                )
            },
        ],
        audit_fieldset_tuple,
    )

    search_fields = (
        "visit_schedule_name",
        "schedule_name",
        "visit_code",
        "visit_title",
        "visit_name",
    )

    def get_list_display(self, request) -> tuple[str, ...]:
        list_display = super().get_list_display(request)
        return (
            "visit_schedule_name",
            "schedule_name",
            "visit_code",
            "visit_title",
            "visit_name",
            "timepoint",
            "active",
            *list_display,
        )

    def get_list_filter(self, request) -> tuple[str, ...]:
        list_filter = super().get_list_filter(request)
        return (
            "active",
            "visit_schedule_name",
            "schedule_name",
            "visit_code",
            *list_filter,
        )

    @staticmethod
    def populate_visit_schedule(request, queryset) -> None:  # noqa: ARG004
        VisitSchedule.objects.update(active=False)
        site_visit_schedules.to_model(VisitSchedule)



from django.db import models

from .field_attrs import get_field_attrs_for_utestid

__all__ = ["result_model_mixin_factory"]


def result_model_mixin_factory(
    utest_id: str,
    units_choices: tuple,
    default_units: str | None = None,
    verbose_name: str | None = None,
    decimal_places: int | None = None,
    max_digits: int | None = None,
    validators: list | None = None,
    help_text: str | None = None,
) -> type[models.Model]:
    """Returns an abstract model class with a single field class"""

    class AbstractModel(models.Model):
        class Meta:
            abstract = True

    for name, fld_cls in get_field_attrs_for_utestid(
        utest_id,
        units_choices,
        default_units,
        verbose_name,
        decimal_places,
        max_digits,
        validators,
        help_text,
    ):
        AbstractModel.add_to_class(name, fld_cls)
    return AbstractModel



from uuid import UUID

from django.apps import apps as django_apps


def create_new_stock_on_receive(receive_item_pk: UUID = None):
    receive_item_model_cls = django_apps.get_model("edc_pharmacy.receiveitem")
    stock_model_cls = django_apps.get_model("edc_pharmacy.stock")
    receive_item = receive_item_model_cls.objects.get(pk=receive_item_pk)
    for i in range(0, int(receive_item.qty)):
        stock_model_cls.objects.create(
            receive_item_id=receive_item.id,
            qty_in=1,
            qty_out=0,
            qty=1,
            product_id=receive_item.order_item.product.id,
            container_id=receive_item.container.id,
            location_id=receive_item.receive.location.id,
            lot_id=receive_item.lot.id,
        )



from django_audit_fields.admin import ModelAdminAuditFieldsMixin
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin

from edc_model_admin.mixins import (
    ModelAdminFormAutoNumberMixin,
    ModelAdminFormInstructionsMixin,
    ModelAdminInstitutionMixin,
    ModelAdminNextUrlRedirectMixin,
    ModelAdminRedirectOnDeleteMixin,
    TemplatesModelAdminMixin,
)
from edc_notification.modeladmin_mixins import NotificationModelAdminMixin


class ModelAdminStackedInlineMixin(
    TemplatesModelAdminMixin,
    ModelAdminNextUrlRedirectMixin,
    NotificationModelAdminMixin,
    ModelAdminFormInstructionsMixin,
    ModelAdminFormAutoNumberMixin,
    ModelAdminRevisionMixin,
    ModelAdminAuditFieldsMixin,
    ModelAdminInstitutionMixin,
    ModelAdminRedirectOnDeleteMixin,
):
    pass



from .fasting_model_mixin_factory import fasting_model_mixin_factory
from .fbg_model_mixin_factory import fbg_model_mixin_factory
from .glucose_model_mixin_factory import glucose_model_mixin_factory
from .ogtt_model_mixin_factory import ogtt_model_mixin_factory



from typing import Any

from edc_dashboard.url_names import url_names
from edc_lab.constants import SHIPPED
from edc_lab.pdf_reports import ManifestPdfReport

from ..listboard_filters import ManifestListboardViewFilters
from .base_listboard_view import BaseListboardView


class ManifestListboardView(BaseListboardView):
    navbar_selected_item = "manifest"

    form_action_url = "manifest_form_action_url"  # url_name
    listboard_url = "manifest_listboard_url"  # url_name
    listboard_template = "manifest_listboard_template"
    listboard_model = "edc_lab.manifest"
    listboard_view_permission_codename = "edc_lab_dashboard.view_lab_manifest_listboard"
    listboard_view_only_my_permission_codename = None
    listboard_view_filters = ManifestListboardViewFilters()
    search_form_url = "manifest_listboard_url"  # url_name
    print_manifest_url = "print_manifest_url"  # url_name

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(
            new_manifest=self.listboard_model_cls(),
            print_manifest_url_name=url_names.get(self.print_manifest_url),
            SHIPPED=SHIPPED,
        )
        return super().get_context_data(**kwargs)

    def get(self, request, *args, **kwargs):
        if request.GET.get("pdf"):
            return self.print_manifest()
        return super().get(request, *args, **kwargs)

    @property
    def manifest(self):
        return self.listboard_model_cls.objects.get(
            manifest_identifier=self.request.GET.get("pdf")
        )

    def print_manifest(self):
        manifest_report = ManifestPdfReport(manifest=self.manifest, user=self.request.user)
        return manifest_report.render_to_response()



from uuid import uuid4

from clinicedc_constants import NORMAL, NULL_STRING
from django.apps import apps as django_apps
from django.conf import settings
from django.contrib.sites.models import Site
from django.db import models
from django.db.models.deletion import PROTECT
from django.template.loader import render_to_string

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_visit_schedule.constants import DAYS, HOURS, MONTHS, WEEKS

from ..choices import QUERY_PRIORITY
from ..site_data_manager import site_data_manager
from .data_dictionary import DataDictionary, DataDictionaryManager
from .query_visit_schedule import QueryVisitSchedule
from .requisition_panel import RequisitionPanel
from .user import DataManagerUser, QueryUser


class QueryRuleError(Exception):
    pass


class QueryRuleManager(models.Manager):
    def get_by_natural_key(self, title):
        return self.get(title=title)


def get_rule_handler_choices(model_name=None):
    choices = []
    for rule_handler in site_data_manager.get_rule_handlers(model_name=model_name):
        choices.append((rule_handler.name, rule_handler.display_name))  # noqa: PERF401
    return tuple(choices) or (
        (DEFAULT_RULE_HANDLER, DEFAULT_RULE_HANDLER.replace("_", " ").title()),
    )


REPORT_DATE = "report_datetime"
DRAWN_DATE = "drawn_datetime"

DATE_CHOICES = (
    (REPORT_DATE, "Report datetime (visit report)"),
    (DRAWN_DATE, "Specimen draw date (requisition)"),
)

UNITS = ((HOURS, "Hours"), (DAYS, "Days"), (WEEKS, "Weeks"), (MONTHS, "Months"))

DEFAULT_RULE_HANDLER = "default"

query_text_template_name = "edc_data_manager/default_query_text.html"


class CrfDataDictionaryManager(DataDictionaryManager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .exclude(
                model__in=[
                    settings.SUBJECT_REQUISITION_MODEL,
                    settings.SUBJECT_VISIT_MODEL,
                ]
            )
        )


class CrfDataDictionary(DataDictionary):
    objects = CrfDataDictionaryManager()

    class Meta:
        proxy = True
        default_permissions = ("view",)


class VisitDataDictionaryManager(DataDictionaryManager):
    def get_queryset(self):
        return super().get_queryset().filter(model=settings.SUBJECT_VISIT_MODEL)


class VisitDataDictionary(DataDictionary):
    objects = VisitDataDictionaryManager()

    class Meta:
        proxy = True
        default_permissions = ("view",)


class RequisitionDataDictionaryManager(DataDictionaryManager):
    def get_queryset(self):
        return super().get_queryset().filter(model=settings.SUBJECT_REQUISITION_MODEL)


class RequisitionDataDictionary(DataDictionary):
    objects = RequisitionDataDictionaryManager()

    class Meta:
        proxy = True
        default_permissions = ("view",)


class QueryRule(BaseUuidModel):
    active = models.BooleanField(default=True)

    title = models.CharField(max_length=150, unique=True)

    reference_model = models.CharField(max_length=150, default=NULL_STRING, editable=False)

    sender = models.ForeignKey(
        DataManagerUser,
        related_name="+",
        on_delete=PROTECT,
        verbose_name="Query raised by",
    )

    visit_schedule = models.ManyToManyField(
        QueryVisitSchedule,
        verbose_name="Visit",
        related_name="+",
        blank=True,
        help_text="select all that apply",
    )

    visit_schedule_exclude = models.ManyToManyField(
        QueryVisitSchedule,
        verbose_name="Visit (exclude)",
        related_name="+",
        blank=True,
        help_text="select all that apply",
    )

    data_dictionaries = models.ManyToManyField(
        CrfDataDictionary,
        verbose_name="Question(s)",
        related_name="+",
        blank=True,
        help_text="select all that apply",
    )

    sites = models.ManyToManyField(Site, help_text="Leave blank to apply to all.")

    requisition_panel = models.ForeignKey(
        RequisitionPanel,
        verbose_name="Are responses linked to a requisition? If so, which",
        related_name="+",
        on_delete=PROTECT,
        null=True,
        blank=True,
        help_text="Requisition will be expected on day of visit.",
    )

    query_text = models.TextField(
        help_text="Generic query text for auto-generated queries.",
        default=NULL_STRING,
        blank=True,
    )

    # TODO: is this the CRF report datetime, for example?
    reference_date = models.CharField(
        max_length=25, choices=DATE_CHOICES, default=REPORT_DATE, editable=False
    )

    # TODO: Does this work? Does this value suggest when to run the ...
    # TODO: ... query relative to the report datetime
    timing = models.IntegerField(verbose_name="Timing", default=48)

    timing_units = models.CharField(
        verbose_name="Units", max_length=25, choices=UNITS, default=HOURS
    )

    query_priority = models.CharField(
        verbose_name="Priority", max_length=25, choices=QUERY_PRIORITY, default=NORMAL
    )

    recipients = models.ManyToManyField(
        QueryUser,
        related_name="+",
        verbose_name="Send to",
        help_text=(
            "Select any additional recipients. Users in the `Site Data Manager` "
            "group are automatically included."
        ),
        blank=True,
    )

    rule_handler_name = models.CharField(
        max_length=150,
        default=DEFAULT_RULE_HANDLER,
        choices=(("do_nothing", "Do Nothing"), ("default", "Default")),
    )

    reference = models.CharField(max_length=36, default=uuid4, unique=True)

    comment = models.TextField(default=NULL_STRING, blank=True)

    objects = QueryRuleManager()

    history = HistoricalRecords()

    def __str__(self):
        inactive = " (inactive)" if not self.active else ""
        return f"{self.title}{inactive}"

    def save(self, *args, **kwargs):
        self.reference_model = settings.SUBJECT_VISIT_MODEL
        super().save(*args, **kwargs)

    def natural_key(self):
        return (self.title,)

    natural_key.dependencies = (
        "edc_data_manager.CrfDataDictionary",
        "edc_data_manager.DataManagerUser",
        "edc_data_manager.QueryUser",
        "edc_data_manager.queryvisitschedule",
        "edc_data_manager.RequisitionPanel",
    )

    @property
    def rendered_query_text(self) -> str:
        context = dict(model=self)
        return render_to_string(template_name=query_text_template_name, context=context)

    @property
    def reference_model_cls(self):
        return django_apps.get_model(self.reference_model)

    @property
    def model_cls(self):
        dct_models = list(set([dd.model for dd in self.data_dictionaries.all()]))
        if len(dct_models) != 1:
            raise QueryRuleError(
                "Multiple models specified. Questions from only one model are allowed."
            )
        return django_apps.get_model(dct_models[0])

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Query Rule"
        verbose_name_plural = "Query Rules"
        indexes = (models.Index(fields=["title", "active"]),)



from .lab import AliquotType

bc = AliquotType(name="Buffy Coat", alpha_code="BC", numeric_code="12")
fbc = AliquotType(name="FBC", alpha_code="FBC", numeric_code="63")
pl = AliquotType(name="Plasma", alpha_code="PL", numeric_code="36")
qfc = AliquotType(name="Quantitative FC", alpha_code="QFC", numeric_code="61")
serum = AliquotType(name="Serum", alpha_code="SERUM", numeric_code="06")
sputum = AliquotType(name="Sputum", alpha_code="SPUTUM", numeric_code="07")
urine = AliquotType(name="Urine", alpha_code="UR", numeric_code="08")
wb = AliquotType(name="Whole Blood", alpha_code="WB", numeric_code="02")

# CSF
csf_store = AliquotType(name="CSF store", alpha_code="CSF", numeric_code="62")
csf_testing = AliquotType(name="Isolates", alpha_code="ISOLATES", numeric_code="64")
csf_glucose = AliquotType(name="Glucose", alpha_code="GLUCOSE", numeric_code="65")
csf_protein = AliquotType(name="Protein", alpha_code="PROTEIN", numeric_code="66")
csf_pellet = AliquotType(name="CSF Pellet", alpha_code="PELLET", numeric_code="67")
csf_supernatant = AliquotType(
    name="CSF Supernatant", alpha_code="SUPERNATANT", numeric_code="68"
)
csf = AliquotType(name="Cerebro Spinal Fluid", alpha_code="CSF", numeric_code="56")
tissue_biopsy = AliquotType(name="Tissue biopsy", alpha_code="TISSUE", numeric_code="44")


# dummy
disposable = AliquotType(name="Disposable", alpha_code="XX", numeric_code="00")


wb.add_derivatives(bc, pl, serum, fbc, wb)
sputum.add_derivatives(sputum)
urine.add_derivatives(urine)



from django.db import models

from edc_model.models import BaseUuidModel

from ...medication import Formulation
from ...stock import Lot
from .container_model_mixin import ContainerModelMixin


class PillBottleError(Exception):
    pass


class PillBottleModelMixin(ContainerModelMixin, BaseUuidModel):
    formulation = models.ForeignKey(Formulation, on_delete=models.PROTECT, blank=True)

    lot = models.ForeignKey(Lot, on_delete=models.PROTECT)

    max_unit_qty = models.IntegerField(blank=False, default=0)

    unit_qty = models.IntegerField(blank=False, default=0)

    source_container = models.ForeignKey(
        "self", on_delete=models.PROTECT, blank=True, null=True
    )

    unit_qty_out = models.IntegerField(default=0)

    def __str__(self):
        return f"{self.formulation} {self.unit_qty} count"

    def save(self, *args, **kwargs):
        self.contains_uniquely_identifiable_items = False
        if self.unit_qty < self.unit_qty_out:
            raise PillBottleError(f"Qty cannot be negative. See {self}.")
        if self.max_unit_qty and self.max_unit_qty > self.unit_qty:
            raise PillBottleError(f"Qty exceeds max_unit_qty. See {self}.")
        self.formulation = self.lot.formulation
        super().save(*args, **kwargs)

    class Meta(ContainerModelMixin.Meta, BaseUuidModel.Meta):
        abstract = True
        verbose_name = "Pill Bottle"
        verbose_name_plural = "Pill Bottles"



from dateutil.relativedelta import relativedelta
from django.db import models

from ..exceptions import RefillEndDatetimeError
from ..refill import adjust_previous_end_datetime


class PreviousNextModelMixin(models.Model):
    def update_or_raise_on_null_refill_end_datetimes(self):
        if self.__class__.objects.filter(refill_end_datetime__isnull=True).exists():
            obj = (
                self.__class__.objects.filter(refill_end_datetime__isnull=True)
                .exclude(id=self.id)
                .order_by("refill_start_datetime")
                .first()
            )
            raise RefillEndDatetimeError(
                f"refill_end_datetime cannot be null. See {obj.__class__.__name__}({obj})."
            )

    @property
    def previous(self):
        # self.update_or_raise_on_null_refill_end_datetimes()
        opts = {"refill_start_datetime__lt": self.refill_start_datetime}
        if getattr(self, "related_visit_model_attr", None):
            opts.update(
                {
                    f"{self.related_visit_model_attr()}__subject_identifier": (
                        self.related_visit.subject_identifier
                    )
                }
            )
        else:
            opts.update({"rx__subject_identifier": self.rx.subject_identifier})
        return (
            self.__class__.objects.filter(**opts)
            .order_by("refill_start_datetime")
            .exclude(refill_identifier=self.refill_identifier)
            .last()
        )

    @property
    def next(self):
        opts = {"refill_start_datetime__gt": self.refill_start_datetime}
        if getattr(self, "related_visit_model_attr", None):
            opts.update(
                {
                    f"{self.related_visit_model_attr()}__subject_identifier": (
                        self.related_visit.subject_identifier
                    )
                }
            )
        else:
            opts.update({"rx__subject_identifier": self.rx.subject_identifier})
        return (
            self.__class__.objects.filter(**opts)
            .order_by("refill_start_datetime")
            .exclude(refill_identifier=self.refill_identifier)
            .first()
        )

    def adjust_end_datetimes(self):
        """Adjust the refill_end_datetime for the previous and
        next instance.

        For example: study medication or rxrefill model instance.
        """
        if self.previous:  # e.g. prev study_med or rx_refill
            adjust_previous_end_datetime(
                self.previous,
                refill_start_datetime=self.refill_start_datetime,
                user_modified=self.user_modified,
                modified=self.modified,
            )
        if self.next:
            self.refill_end_datetime = self.next.refill_start_datetime - relativedelta(
                minutes=1
            )

    class Meta:
        abstract = True



# Generated by Django 5.2.6 on 2025-09-17 16:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_action_item",
            "0038_alter_actionitem_revision_alter_actiontype_revision_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="actionitem",
            name="auto_created_comment",
            field=models.CharField(blank=True, default="", max_length=25),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="instructions",
            field=models.TextField(
                blank=True, default="", help_text="populated by action class"
            ),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="May be left blank. e.g. action identifier from reference model that opened the item (parent).",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="priority",
            field=models.CharField(
                blank=True,
                choices=[
                    ("high", "High Priority"),
                    ("medium", "Medium Priority"),
                    ("low", "Low Priority"),
                ],
                default="",
                help_text="Leave blank to use default for this action type.",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="May be left blank. e.g. action identifier from source model that opened the item.",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="actiontype",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="actiontype",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="actiontype",
            name="instructions",
            field=models.TextField(blank=True, default="", max_length=250),
        ),
        migrations.AlterField(
            model_name="actiontype",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="actiontype",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="actiontype",
            name="reference_model",
            field=models.CharField(
                blank=True, default="", help_text="reference model", max_length=100
            ),
        ),
        migrations.AlterField(
            model_name="actiontype",
            name="related_reference_model",
            field=models.CharField(
                blank=True,
                default="",
                help_text="origin reference model",
                max_length=100,
            ),
        ),
        migrations.AlterField(
            model_name="edcpermissions",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="edcpermissions",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="edcpermissions",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="edcpermissions",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="auto_created_comment",
            field=models.CharField(blank=True, default="", max_length=25),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="instructions",
            field=models.TextField(
                blank=True, default="", help_text="populated by action class"
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="May be left blank. e.g. action identifier from reference model that opened the item (parent).",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="priority",
            field=models.CharField(
                blank=True,
                choices=[
                    ("high", "High Priority"),
                    ("medium", "Medium Priority"),
                    ("low", "Low Priority"),
                ],
                default="",
                help_text="Leave blank to use default for this action type.",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="May be left blank. e.g. action identifier from source model that opened the item.",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalreference",
            name="action_item_reason",
            field=models.TextField(default="", editable=False),
        ),
        migrations.AlterField(
            model_name="historicalreference",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalreference",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalreference",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreference",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreference",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to parent reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="historicalreference",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to related reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="action_item_reason",
            field=models.TextField(default="", editable=False),
        ),
        migrations.AlterField(
            model_name="reference",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to parent reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to related reference model instance.",
                max_length=30,
            ),
        ),
    ]



from __future__ import annotations

from django.apps import apps as django_apps
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.db.models import QuerySet
from django.http import FileResponse, HttpResponseRedirect
from django.urls import reverse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.views.generic.base import TemplateView
from pylabels import Sheet, Specification

from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin
from edc_protocol.view_mixins import EdcProtocolViewMixin
from edc_pylabels.models import LabelConfiguration
from edc_pylabels.site_label_configs import site_label_configs

from ..models import Stock


@method_decorator(login_required, name="dispatch")
class PrintLabelsView(EdcViewMixin, NavbarViewMixin, EdcProtocolViewMixin, TemplateView):
    stock_pks: list[str] | None = None
    template_name: str = "edc_pharmacy/stock/print_labels.html"
    session_key = "model_pks"
    navbar_name = settings.APP_NAME
    navbar_selected_item = "pharmacy"
    show_watermark = True

    def get_context_data(self, **kwargs):
        session_uuid = str(kwargs.get("session_uuid"))
        selected_label_configuration = kwargs.get("label_configuration")
        stock_pks = self.request.session.get(session_uuid, [])
        try:
            _, querystring = self.request.META.get("HTTP_REFERER").split("?")
        except ValueError:
            querystring = ""
        kwargs.update(
            source_changelist_url=self.source_changelist_url,
            source_model_name=Stock._meta.verbose_name,
            label_configurations=LabelConfiguration.objects.all(),
            selected_label_configuration=selected_label_configuration,
            q=querystring,
            max_to_print=len(stock_pks),
        )
        return super().get_context_data(**kwargs)

    @property
    def source_changelist_url(self):
        return reverse(
            f"edc_pharmacy_admin:edc_pharmacy_{self.kwargs.get('model')}_changelist"
        )

    @property
    def model_cls(self):
        return django_apps.get_model(f"edc_pharmacy.{self.kwargs.get('model')}")

    def post(self, request, *args, **kwargs):  # noqa: ARG002
        session_uuid = str(kwargs.get("session_uuid"))
        stock_pks = request.session.get(session_uuid, [])
        url = self.source_changelist_url
        if stock_pks:
            opts = {}
            filter_by = self.request.POST.get("filter_by")
            if filter_by == "confirmed_only":
                opts = dict(confirmation__isnull=False)
            elif filter_by == "unconfirmed_only":
                opts = dict(confirmation__isnull=True)
            queryset: QuerySet[Stock] = self.model_cls.objects.filter(
                pk__in=stock_pks, **opts
            ).order_by("code")
            max_to_print = int(self.request.POST.get("max_to_print"), 0)
            if 0 < max_to_print < len(stock_pks):
                queryset = queryset[0:max_to_print]

            del request.session[session_uuid]

            label_configuration: LabelConfiguration = LabelConfiguration.objects.get(
                pk=request.POST.get("label_configuration")
            )
            if label_configuration.requires_allocation and queryset.filter(
                allocation__isnull=True
            ):
                messages.add_message(
                    request,
                    messages.ERROR,
                    (
                        "Unable to print selected stock items. "
                        f"Label format '{label_configuration.name}' may only be used with "
                        "stock items allocated to subjects."
                    ),
                )
                return HttpResponseRedirect(request.META.get("HTTP_REFERER", ""))
            label_data = [obj for obj in queryset]
            drawing_callable = site_label_configs.get(
                label_configuration.name
            ).drawing_callable
            specs = Specification(**label_configuration.label_specification.as_dict)
            sheet = Sheet(
                specs,
                drawing_callable,
                border=label_configuration.label_specification.border,
            )
            try:
                sheet.add_labels(label_data)
            except AttributeError:
                messages.add_message(
                    request,
                    messages.ERROR,
                    (
                        "Unable to print these items using the "
                        f"label format '{label_configuration.name}'. Perhaps try another."
                    ),
                )
                return HttpResponseRedirect(request.META.get("HTTP_REFERER", ""))
            else:
                buffer = sheet.save_to_buffer()
                now = timezone.now()
                return FileResponse(
                    buffer,
                    as_attachment=True,
                    filename=(
                        f"{label_configuration.name}_{now.strftime('%Y-%m-%d %H:%M')}.pdf"
                    ),
                )
        return HttpResponseRedirect(url)



import sys

from django.core.management.base import BaseCommand
from tqdm import tqdm

from edc_appointment.models import Appointment
from edc_appointment.utils import reset_visit_code_sequence_or_pass
from edc_registration.models import RegisteredSubject


class Command(BaseCommand):
    help = (
        "Validate appointment visit code sequences relative to appt_datetime "
        "and reset if needed."
    )

    def handle(self, *args, **options):  # noqa: ARG002
        sys.stdout.write(
            "Validating (and resetting, if needed) appointment visit code sequences ...\n"
        )
        qs_rs = RegisteredSubject.objects.all().order_by("subject_identifier")
        for obj in tqdm(qs_rs, total=qs_rs.count()):
            qs = Appointment.objects.filter(
                subject_identifier=obj.subject_identifier,
                visit_code_sequence=0,
            ).order_by("subject_identifier", "visit_code")
            for appointment in qs:
                reset_visit_code_sequence_or_pass(
                    subject_identifier=appointment.subject_identifier,
                    visit_schedule_name=appointment.visit_schedule_name,
                    schedule_name=appointment.schedule_name,
                    visit_code=appointment.visit_code,
                    write_stdout=True,
                )



from __future__ import annotations

from clinicedc_constants import CONFIRMED
from django.apps import apps as django_apps
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from django.views.generic.base import TemplateView

from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin
from edc_protocol.view_mixins import EdcProtocolViewMixin

from ..constants import ALREADY_CONFIRMED, INVALID
from ..models import Stock
from ..utils import confirm_stock


@method_decorator(login_required, name="dispatch")
class ConfirmStockFromQuerySetView(
    EdcViewMixin, NavbarViewMixin, EdcProtocolViewMixin, TemplateView
):
    template_name: str = "edc_pharmacy/stock/confirm_stock_by_queryset.html"
    navbar_name = settings.APP_NAME
    navbar_selected_item = "pharmacy"
    codes_per_page = 12

    def get_context_data(self, **kwargs):
        kwargs.update(
            CONFIRMED=CONFIRMED,
            ALREADY_CONFIRMED=ALREADY_CONFIRMED,
            INVALID=INVALID,
            item_count=list(range(1, self.adjusted_unconfirmed_count + 1)),
            unconfirmed_count=self.unconfirmed_count,
            confirmed_count=self.confirmed_count,
            source_changelist_url=self.source_changelist_url,
            **self.session_data,
        )
        return super().get_context_data(**kwargs)

    @property
    def session_data(self):
        session_uuid = self.kwargs.get("session_uuid")
        session_data = {}
        if session_uuid:
            session_obj = self.request.session[str(session_uuid)]
            transaction_word = _(session_obj.get("transaction_word") or "confirmed")
            last_stock_codes = [
                [x, CONFIRMED, transaction_word]
                for x in session_obj.get("confirmed_codes") or []
            ]
            last_stock_codes.extend(
                [
                    [
                        x,
                        ALREADY_CONFIRMED,
                        _("already %(transaction_word)s")
                        % {"transaction_word": transaction_word},
                    ]
                    for x in session_obj.get("already_confirmed_codes") or []
                ]
            )
            last_stock_codes.extend(
                [[x, INVALID, _("invalid")] for x in session_obj.get("invalid_codes") or []]
            )
            session_data = dict(
                transaction_word=session_obj.get("transaction_word"),
                source_pk=session_obj.get("source_pk"),
                source_identifier=session_obj.get("source_identifier"),
                source_label_lower=session_obj.get("source_label_lower"),
                source_model_name=session_obj.get("source_model_name"),
                last_stock_codes=last_stock_codes,
                stock_codes=session_obj.get("stock_codes") or [],
                total_stock_code_count=len(session_obj.get("stock_codes")),
            )
        return session_data

    @property
    def unconfirmed_count(self) -> int:
        return (
            Stock.objects.values("code")
            .filter(
                code__in=self.session_data.get("stock_codes"),
                confirmation__isnull=True,
            )
            .count()
        )

    @property
    def confirmed_count(self) -> int:
        return (
            Stock.objects.values("code")
            .filter(
                code__in=self.session_data.get("stock_codes"),
                confirmation__isnull=False,
            )
            .count()
        )

    @property
    def adjusted_unconfirmed_count(self) -> int:
        return min(self.unconfirmed_count, self.codes_per_page)

    @property
    def source_changelist_url(self):
        model = self.session_data.get("source_label_lower").split(".")[1]
        return reverse(f"edc_pharmacy_admin:edc_pharmacy_{model}_changelist")

    @property
    def source_model_cls(self):
        label_lower = self.session_data.get("source_label_lower")
        return django_apps.get_model(label_lower)

    def post(self, request, *args, **kwargs):  # noqa: ARG002
        stock_codes = request.POST.getlist("stock_codes")
        if len(stock_codes) != len(list(set(stock_codes))):
            messages.add_message(
                request,
                messages.ERROR,
                (
                    "Stock items not confirmed. List of stock codes "
                    "is not unique. Please check and scan again."
                ),
            )
            url = reverse("edc_pharmacy:confirm_stock_from_queryset_url", kwargs=kwargs)
        else:
            confirmed_codes, already_confirmed_codes, invalid_codes = confirm_stock(
                None,
                stock_codes,
                None,
                confirmed_by=request.user.username,
                user_created=request.user.username,
            )
            if confirmed_codes:
                messages.add_message(
                    request,
                    messages.SUCCESS,
                    f"Successfully confirmed {len(confirmed_codes)} stock items. ",
                )
            if already_confirmed_codes:
                messages.add_message(
                    request,
                    messages.WARNING,
                    (
                        f"Skipped {len(already_confirmed_codes)} items. Stock items are "
                        "already confirmed."
                    ),
                )
            if invalid_codes:
                messages.add_message(
                    request,
                    messages.ERROR,
                    f"Invalid codes submitted! Got {', '.join(invalid_codes)} .",
                )

            if (
                Stock.objects.values("code")
                .filter(
                    code__in=self.session_data.get("stock_codes"),
                    confirmation__isnull=True,
                )
                .exists()
            ):
                self.request.session[str(self.kwargs.get("session_uuid"))] = dict(
                    confirmed_codes=confirmed_codes,
                    already_confirmed_codes=already_confirmed_codes,
                    invalid_codes=invalid_codes,
                    source_pk=self.session_data.get("source_pk"),
                    source_identifier=self.session_data.get("source_identifier"),
                    source_label_lower=self.session_data.get("source_label_lower"),
                    source_model_name=self.session_data.get("source_model_name"),
                    stock_codes=self.session_data.get("stock_codes"),
                )
                url = reverse("edc_pharmacy:confirm_stock_from_queryset_url", kwargs=kwargs)
            else:
                self.request.session[self.kwargs.get("session_uuid")] = None
                url = f"{self.source_changelist_url}?q="
        return HttpResponseRedirect(url)



from __future__ import annotations

from django.contrib import admin, messages
from django.db.models import QuerySet
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext

from ...models import StorageBin


@admin.display(description="Add stock to storage bin")
def add_to_storage_bin_action(modeladmin, request, queryset: QuerySet[StorageBin]):
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            gettext("Select one and only one item"),
        )
    else:
        url = reverse(
            "edc_pharmacy:add_to_storage_bin_url",
            kwargs={"storage_bin": queryset.first().id},
        )
        return HttpResponseRedirect(url)
    return None


@admin.display(description="Move stock to selected storage bin")
def move_to_storage_bin_action(modeladmin, request, queryset: QuerySet[StorageBin]):
    url = reverse(
        "edc_pharmacy:move_to_storage_bin_url",
        kwargs={"storage_bin": queryset.first().id},
    )
    return HttpResponseRedirect(url)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist

from ..single_site import SiteDomainRequiredError

if TYPE_CHECKING:
    from ..single_site import SingleSite


def get_or_create_site_obj(single_site: SingleSite, apps):
    if "multisite" in settings.INSTALLED_APPS and not single_site.domain:
        raise SiteDomainRequiredError(
            f"Domain required when using `multisite`. Got None for `{single_site.name}`"
        )
    site_model_cls = apps.get_model("sites", "Site")
    try:
        site_obj = site_model_cls.objects.get(pk=single_site.site_id)
    except ObjectDoesNotExist:
        site_obj = site_model_cls.objects.create(
            pk=single_site.site_id, name=single_site.name, domain=single_site.domain
        )
    else:
        site_obj.name = single_site.name
        site_obj.domain = single_site.domain
        site_obj.save()
    return site_obj



from django.core.management.base import BaseCommand, CommandError
from django.core.management.color import color_style

from ...update_metadata_on_schedule_change import (
    UpdateMetadataError,
    UpdateMetadataOnScheduleChange,
)

style = color_style()


class Command(BaseCommand):
    help = "Update metadata for changed visit_schedule/schedule names"
    pattern = "^[0-9a-z_]+$"
    fieldnames = ("visit_schedule_name", "schedule_name")

    def add_arguments(self, parser):
        parser.add_argument(
            "--field",
            dest="field",
            default=None,
            help='Field name. Either "visit_schedule_name" or "schedule_name"',
        )

        parser.add_argument(
            "--old-value",
            dest="old_value",
            default=None,
            help="Old or existing value",
        )

        parser.add_argument(
            "--new-value",
            dest="new_value",
            default=None,
            help="New value to replace the old value",
        )

        parser.add_argument(
            "--dry-run",
            dest="dry_run",
            default=True,
            help="Do a dry run. (Default: True)",
        )

    def handle(self, *args, **options):  # noqa: ARG002
        dry_run = options.get("dry_run", "") != "False"

        try:
            UpdateMetadataOnScheduleChange(
                field=options.get("field"),
                new_name=options.get("new_value"),
                old_name=options.get("old_value"),
                dry_run=dry_run,
            )
        except UpdateMetadataError as e:
            raise CommandError(e) from e



from django.contrib.auth.mixins import LoginRequiredMixin
from django.http.response import HttpResponseRedirect
from django.urls.base import reverse
from django.views.generic.base import View

from edc_dashboard.url_names import url_names
from edc_label.printers_mixin import PrintersMixin


class BaseRequisitionView(LoginRequiredMixin, PrintersMixin, View):
    success_url_name = "subject_dashboard_url"
    lab_home_url_name = "requisition_listboard_url"

    def get_success_url(self):
        return url_names.get(self.success_url_name)

    def get(self, request, *args, **kwargs):  # noqa: ARG002
        url = reverse(url_names.get(self.lab_home_url_name))
        return HttpResponseRedirect(url)

    def head(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def options(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)



# Generated by Django 4.2.1 on 2023-07-07 19:32

from django.db import migrations

import edc_lab.models.aliquot
import edc_lab.models.box
import edc_lab.models.manifest.manifest
import edc_lab.models.order
import edc_lab.models.result
import edc_lab.models.result_item
import edc_sites.models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_lab", "0026_alter_aliquot_options_alter_box_options_and_more"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="aliquot",
            managers=[
                ("objects", edc_lab.models.aliquot.Manager()),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="box",
            managers=[
                ("objects", edc_lab.models.box.BoxManager()),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="manifest",
            managers=[
                ("objects", edc_lab.models.manifest.manifest.Manager()),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="order",
            managers=[
                ("objects", edc_lab.models.order.OrderManager()),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="result",
            managers=[
                ("objects", edc_lab.models.result.ResultManager()),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="resultitem",
            managers=[
                ("objects", edc_lab.models.result_item.ResultItemManager()),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
    ]



import pandas as pd
from clinicedc_constants import FEMALE, MALE

from .constants import COUNT_COLUMN, N_ONLY, N_WITH_ROW_PROP, TITLE_COLUMN
from .row import RowDefinition, RowDefinitions, RowStatisticsWithGender


class Table:

    title_column = "Characteristics"
    label_column = "Statistic"
    default_sublabel = "n"
    gender_column = "gender"
    row_statistics_cls: RowStatisticsWithGender = RowStatisticsWithGender

    def __init__(
        self,
        colname: str | None = None,
        main_df: pd.DataFrame = None,
        title: str | None = None,
        include_zero_counts: bool | None = None,
    ):

        self.colname = colname
        self.main_df = main_df
        self.title = title
        self.include_zero_counts = include_zero_counts
        self.table_df: pd.DataFrame = pd.DataFrame()

        self.build_table_df()
        if self.title:
            # add a redundant column to hold title name for each
            # row in this table.
            self.table_df[TITLE_COLUMN] = self.title

    @property
    def row_definitions(self) -> RowDefinitions:
        """Override with your RowDefs

        The default adds a first row with gender breakdown.
        """
        row_defs = RowDefinitions(
            colname=self.colname, row_statistics_cls=self.row_statistics_cls
        )
        row_defs.add(
            RowDefinition(
                title=self.title,
                label=self.default_sublabel,
                colname=None,
                condition=(self.main_df[self.gender_column].notna()),
                columns={
                    FEMALE: (N_WITH_ROW_PROP, 2),
                    MALE: (N_WITH_ROW_PROP, 2),
                    "All": (N_ONLY, 2),
                },
                drop=False,
            )
        )
        return row_defs

    def reorder_df(self):
        """Override to reorder the rows in `table_df`."""
        pass

    def build_table_df(self) -> None:
        """Build the table_df using the row definitions."""
        df_denominator = self.main_df.copy()
        rows = []
        for index, rd in enumerate(self.row_definitions.definitions):
            if not rd.condition.empty:
                df_numerator = df_denominator.loc[rd.condition]
            else:
                # default to first col non-null values
                df_numerator = df_denominator.loc[
                    df_denominator[df_denominator.columns[0]].notna()
                ]
            row_stats = self.row_statistics_cls(
                colname=rd.colname,
                df_numerator=df_numerator,
                df_denominator=df_denominator,
                df_all=self.main_df,
                columns=rd.columns,
            )
            if index == 0:
                columns = (
                    [self.title_column, self.label_column]
                    + row_stats.labels()
                    + [TITLE_COLUMN]
                )
                # reset table_df
                self.table_df = pd.DataFrame(columns=columns)
            rows.append([rd.title, rd.label] + row_stats.values_list() + [self.title])
            if rd.drop and not df_numerator.empty:
                df_denominator.drop(df_numerator.index, inplace=True)
        if self.row_definitions.reverse_rows:
            rows.reverse()
        for index, values_list in enumerate(rows):
            self.table_df.loc[index] = values_list
        if not self.include_zero_counts:
            self.table_df.drop(
                self.table_df[self.table_df[COUNT_COLUMN] == 0].index, inplace=True
            )
        self.reorder_df()

    @property
    def formatted_df(self) -> pd.DataFrame:
        """Return DF with first 5 columns"""
        return self.table_df.iloc[:, :5]



from edc_identifier.simple_identifier import SimpleUniqueIdentifier


class ScreeningIdentifier(SimpleUniqueIdentifier):
    # TODO: is this connected to edc_protocol.screening_identifier_pattern ??
    random_string_length = 7
    identifier_type = "screening_identifier"
    template = "S{device_id}{random_string}"



# Generated by Django 5.0.7 on 2024-08-12 16:18

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_qareports", "0009_auto_20240812_1918"),
    ]

    operations = []
